@@ -4,6 +4,7 @@ use self::Term::*;
 use self::Error::*;
 use std::fmt;
 use std::borrow::Cow;
+use std::char::from_u32;
 
 /// A lambda term that is either a variable with a De Bruijn index, an abstraction over a term or
 /// an applicaction of one term to another.
@@ -231,40 +232,6 @@ impl Term {
     pub fn rhs_ref_mut(&mut self) -> Result<&mut Term, Error> {
         if let Ok((_, rhs)) = self.unapp_ref_mut() { Ok(rhs) } else { Err(NotAnApp) }
     }
-
-    /// Applies `self` to another term with substitution and variable update, consuming them
-    /// in the process.
-    ///
-    /// # Example
-    /// ```
-    /// use lambda_calculus::parser::parse;
-    ///
-    /// let lhs    = parse(&"λλ42(λ13)").unwrap();
-    /// let rhs    = parse(&"λ51").unwrap();
-    /// let result = parse(&"λ3(λ61)(λ1(λ71))").unwrap();
-    ///
-    /// assert_eq!(lhs.apply(rhs), Ok(result));
-    /// ```
-    pub fn apply(self, rhs: Term) -> Result<Term, Error> {
-        apply(self, rhs)
-    }
-
-    /// Reduces an `App`lication by substitution and variable update.
-    ///
-    /// # Example
-    /// ```
-    /// use lambda_calculus::term::{app, abs};
-    /// use lambda_calculus::term::Term::Var;
-    /// use lambda_calculus::arithmetic::zero;
-    /// use lambda_calculus::combinators::i;
-    ///
-    /// assert_eq!(app(i(), zero()).eval(), Ok(abs(abs(Var(1)))));
-    /// ```
-    pub fn eval(self) -> Result<Term, Error> {
-        let (lhs, rhs) = try!(self.unapp());
-
-        apply(lhs, rhs)
-    }
 }
 
 /// Wraps a term in an abstraction. Consumes its argument.
@@ -290,82 +257,47 @@ pub fn abs(term: Term) -> Term { Abs(Box::new(term)) }
 /// ```
 pub fn app(lhs: Term, rhs: Term) -> Term { App(Box::new(lhs), Box::new(rhs)) }
 
-/// Applies two terms with substitution and variable update, consuming them in the process.
-///
-/// # Example
-/// ```
-/// use lambda_calculus::term::apply;
-/// use lambda_calculus::parser::parse;
-///
-/// let lhs    = parse(&"λλ42(λ13)").unwrap();
-/// let rhs    = parse(&"λ51").unwrap();
-/// let result = parse(&"λ3(λ61)(λ1(λ71))").unwrap();
-///
-/// assert_eq!(apply(lhs, rhs), Ok(result));
-/// ```
-pub fn apply(mut lhs: Term, rhs: Term) -> Result<Term, Error> {
-    _apply(&mut lhs, rhs, 0);
-
-    lhs.unabs()
-}
-
-fn _apply(lhs: &mut Term, rhs: Term, depth: usize) {
-    match *lhs {
-        Var(i) => if i == depth {
-            *lhs = rhs; // substitute a top-level variable from lhs with rhs
-            update_free_variables(lhs, depth - 1, 0); // update indices of free variables from rhs
-        } else if i > depth {
-            *lhs = Var(i - 1) // decrement a free variable's index
-        },
-        Abs(_) => {
-            _apply(lhs.unabs_ref_mut().unwrap(), rhs, depth + 1)
-        },
-        App(_, _) => {
-            _apply(lhs.lhs_ref_mut().unwrap(), rhs.clone(), depth);
-            _apply(lhs.rhs_ref_mut().unwrap(), rhs, depth)
-        }
-    }
-}
-
-fn update_free_variables(term: &mut Term, added_depth: usize, own_depth: usize) {
-    match *term {
-        Var(i) => if i > own_depth {
-            *term = Var(i + added_depth)
-        },
-        Abs(_) => {
-            update_free_variables(term.unabs_ref_mut().unwrap(), added_depth, own_depth + 1)
-        },
-        App(_, _) => {
-            update_free_variables(term.lhs_ref_mut().unwrap(), added_depth, own_depth);
-            update_free_variables(term.rhs_ref_mut().unwrap(), added_depth, own_depth)
-        }
-    }
-}
-
 /// Set to `true` for λ or `false` for \ when displaying lambda terms. The default is `true`.
 pub const DISPLAY_PRETTY: bool = true;
 
+/// Set to `true` for classic lambda mode or `false` for the De Bruijn index display mode when
+/// displaying lambda terms. The default is `false`.
+pub const DISPLAY_CLASSIC: bool = false;
+
 impl fmt::Display for Term {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}", show_precedence(0, self))
+        write!(f, "{}", show_precedence(self, 0, 0))
     }
 }
 
-fn show_precedence(context_precedence: usize, term: &Term) -> String {
+// TODO: make it safer (unwraps)
+fn show_precedence(term: &Term, context_precedence: usize, depth: u32) -> String {
     match *term {
-        Var(i) => format!("{:X}", i), // max. index = 15
+        Var(i) => if DISPLAY_CLASSIC {
+            format!("{}", from_u32(depth + 97 - i as u32).unwrap())
+        } else {
+            format!("{:X}", i)
+        },
         Abs(ref t) => {
-            let ret = format!("{}{}", if DISPLAY_PRETTY { 'λ' } else { '\\' }, t);
-            parenthesize_if(context_precedence > 1, &ret).into()
+            let ret = if DISPLAY_CLASSIC {
+                format!("{}{}. {}", if DISPLAY_PRETTY { 'λ' } else { '\\' }, from_u32(depth + 97).unwrap(), show_precedence(t, 0, depth + 1))
+            } else {
+                format!("{}{}", if DISPLAY_PRETTY { 'λ' } else { '\\' }, t)
+            };
+            parenthesize_if(&ret, context_precedence > 1).into()
         },
         App(ref t1, ref t2) => {
-            let ret = format!("{}{}", show_precedence(2, t1), show_precedence(3, t2));
-            parenthesize_if(context_precedence == 3, &ret).into()
+            let ret = if DISPLAY_CLASSIC {
+                format!("{} {}", show_precedence(t1, 2, depth), show_precedence(t2, 3, depth))
+            } else {
+                format!("{}{}", show_precedence(t1, 2, depth), show_precedence(t2, 3, depth))
+            };
+            parenthesize_if(&ret, context_precedence == 3).into()
         }
     }
 }
 
-fn parenthesize_if(condition: bool, input: &str) -> Cow<str> {
+fn parenthesize_if(input: &str, condition: bool) -> Cow<str> {
     if condition {
         format!("({})", input).into()
     } else {
@@ -375,26 +307,19 @@ fn parenthesize_if(condition: bool, input: &str) -> Cow<str> {
 
 #[cfg(test)]
 mod test {
+    use super::DISPLAY_CLASSIC;
     use arithmetic::{zero, succ, pred};
-    use combinators::i;
-    use parser::parse;
-    use super::{apply, abs};
-    use super::Term::Var;
 
     #[test]
-    fn applying() {
-        let lhs    = parse(&"λλ42(λ13)").unwrap();
-        let rhs    = parse(&"λ51").unwrap();
-        let result = parse(&"λ3(λ61)(λ1(λ71))").unwrap();
-        assert_eq!(apply(lhs, rhs), Ok(result));
-
-        assert_eq!(i().app(zero()).eval().unwrap(), abs(abs(Var(1))));
-    }
-
-    #[test]
-    fn displaying_terms() {
-        assert_eq!(&format!("{}", zero()), "λλ1");
-        assert_eq!(&format!("{}", succ()), "λλλ2(321)");
-        assert_eq!(&format!("{}", pred()), "λλλ3(λλ1(24))(λ2)(λ1)");
+    fn display_modes() {
+        if DISPLAY_CLASSIC {
+            assert_eq!(&format!("{}", zero()), "λa. λb. b");
+            assert_eq!(&format!("{}", succ()), "λa. λb. λc. b (a b c)");
+            assert_eq!(&format!("{}", pred()), "λa. λb. λc. a (λd. λe. e (d b)) (λd. c) (λd. d)");
+        } else {
+            assert_eq!(&format!("{}", zero()), "λλ1");
+            assert_eq!(&format!("{}", succ()), "λλλ2(321)");
+            assert_eq!(&format!("{}", pred()), "λλλ3(λλ1(24))(λ2)(λ1)");
+        }
     }
 }
